syntax = "proto2";
option optimize_for = LITE_RUNTIME;

package mir.protobuf.wire;

message Invocation {
  required uint32 id = 1;
  required string method_name = 2;
  required bytes  parameters = 3;
  required uint32 protocol_version = 4;
  optional uint32 side_channel_fds = 5;
}

message Result {
  // Invocation results have id and response.
  optional uint32 id = 1;
  optional bytes response = 2;
  // Events are in events.
  repeated bytes events = 3;
}
