/*
 * Copyright © 2019 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: William Wold <william.wold@canonical.com>
 */

#ifndef MIR_FRONTEND_XDG_OUTPUT_V1_H
#define MIR_FRONTEND_XDG_OUTPUT_V1_H

#include <memory>

struct wl_display;

namespace mir
{
namespace frontend
{
class XdgOutputManagerV1;
class OutputManager;

auto create_xdg_output_manager_v1(struct wl_display* display, OutputManager* const output_manager)
    -> std::shared_ptr<XdgOutputManagerV1>;

}
}

#endif // MIR_FRONTEND_XDG_OUTPUT_V1_H
