body {
  background-color: #fff;
  color: #333;
  font-family: Ubuntu,Arial,sans-serif;
  font-size: 10pt;
  margin-right: 100px;
  margin-left: 100px;
}

h1, h2, h3, h4, h5, h6, .toc_title {
  color: #dd4814;
  margin-top: 2em;
  margin-bottom: 1em;
}

h1 {
  text-align: center;
  font-size: 18pt;
}

h2, .toc_title {
  font-weight: bold;
  font-size: 12pt;
  margin-left: -40px;
}

h3, h4, h5, h6 {
  font-size: 10pt;
  margin-left: -20px;
}

.toc_category, .toc_stylepoint {
  font-size: 10pt;
  padding-top: .3em;
  padding-bottom: .3em;
}

table {
  border-collapse: collapse;
}

td, th {
  border: 1px solid #ccc;
  padding: 2px 12px;
  font-size: 10pt;
}

.toc td, .toc th {
  border-width: 1px 5px;
}

code, samp, var {
  color: #060;
}

pre {
  font-size: 10pt;
  display: block;
  color: #060;
  background-color: #f8fff8;
  border-color: #f0fff0;
  border-style: solid;
  border-top-width: 1px;
  border-bottom-width: 1px;
  border-right-width: 1px;
  border-left-width: 5px;
  padding-left: 12px;
  padding-right: 12px;
  padding-top: 4px;
  padding-bottom: 4px;
}

pre.badcode {
  color: #c00;
  background-color: #fff8f8;
  border-color: #fff0f0;
}

.showhide_button {
  float: left;
  cursor: pointer;
  border-width: 1px;
  border-style: solid;
  border-color: #ddd #aaa #aaa #ddd;
  padding: 0 3px 1px;
  margin: 0 4px 8px 0;
  border-radius: 3px;
  -webkit-border-radius: 3px;
  -moz-border-radius: 3px;
}

.link_button {
  float: left;
  display: none;
  background-color: #f8f8ff;
  border-color: #f0f0ff;
  border-style: solid;
  border-width: 1px;
  font-size: 75%;
  margin-top: 0;
  margin-left: -50px;
  padding: 4px;
  border-radius: 3px;
  -webkit-border-radius: 3px;
  -moz-border-radius: 3px;
}

address {
  text-align: right;
}

hr {
  margin-top: 3.5em;
  border-width: 1px;
  color: #fff;
}

.stylepoint_section {
  display: block;
  margin-bottom: 1em;
  color: #772953;
  font-family: Ubuntu,Arial,sans-serif;
  font-size: 90%;
  font-weight: bold;
  margin-left: -2%;
}

.stylepoint_subsection {
  color: #667799;
  font-family: Ubuntu,Arial,sans-serif;
  font-size: 90%;
  font-weight: bold;
  margin-left: -1%;
}

.stylepoint_subsubsection {
  color: #667799;
  font-family: Ubuntu,Arial,sans-serif;
  font-size: 80%;
  font-weight: bold;
  margin-left: 0;
}

.revision {
  text-align: right;
}

