/* The probes need to be compiled in a C file (not C++) */
#define TRACEPOINT_CREATE_PROBES

#include "compositor_report_tp.h"
#include "input_report_tp.h"
#include "connector_report_tp.h"
#include "display_report_tp.h"
#include "session_mediator_report_tp.h"
#include "scene_report_tp.h"
#include "message_processor_report_tp.h"
#include "shared_library_prober_report_tp.h"
