/*
 * Copyright © 2016 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 or 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Alan Griffiths <alan@octopull.co.uk>
 */

#include "miral/add_init_callback.h"

#include <mir/server.h>

miral::AddInitCallback::AddInitCallback(Callback const& callback) :
    callback{callback}
{
}

miral::AddInitCallback::~AddInitCallback() = default;

void miral::AddInitCallback::operator()(mir::Server& server) const
{
    server.add_init_callback(callback);
}
