/*
 * Copyright © 2012-2014 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License version 2 or 3,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef MIR_CLIENT_LIBRARY_H
#define MIR_CLIENT_LIBRARY_H

#include <mir_toolkit/mir_connection.h>
#include <mir_toolkit/mir_window.h>
#include <mir_toolkit/mir_wait.h>
#include <mir_toolkit/mir_presentation_chain.h>
#include <mir_toolkit/rs/mir_render_surface.h>
#include <mir_toolkit/mir_buffer.h>
#include <mir_toolkit/mir_buffer_stream.h>
#include <mir_toolkit/version.h>
#include <mir_toolkit/cursors.h>
#include <mir_toolkit/mir_cookie.h>
#include <mir_toolkit/mir_display_configuration.h>
#include <mir_toolkit/mir_input_device.h>
#include <mir_toolkit/mir_error.h>
#include <mir_toolkit/mir_window_id.h>

#endif /* MIR_CLIENT_LIBRARY_H */
